# flake8: noqa

from starlette_web.common.database import ModelBase


"""
Initialize your models in this file.
It will be automatically introspected, 
if you add %app_name% from "apps.py" to settings.INSTALLED_APPS

>>> class Model(ModelBase):
>>>     __tablename__ = 'app_model'
>>>     ...
"""
